;------------------------------------------------------------------------------------------
; Neural Network Object
; (C) 2002 by Exagone [Thomas Bleeker]. http://www.MadWizard.org
;
;
;------------------------------------------------------------------------------------------

; --=====================================================================================--
; CLASS METHOD PROTOS
; --=====================================================================================--
NNet_Init    			PROTO  :DWORD, :DWORD

; --=====================================================================================--
; FUNCTION POINTER PROTOS
; --=====================================================================================--
NNet_destructorPto   			TYPEDEF  PROTO  :DWORD 
NNet_trainPto					TYPEDEF	 PROTO	:DWORD, :DWORD, :DWORD
NNet_initRandomPto				TYPEDEF	 PROTO	:DWORD
NNet_calcOutputsPto				TYPEDEF  PROTO  :DWORD
NNet_calcOutputDeltasPto		TYPEDEF	 PROTO	:DWORD, :DWORD
NNet_setLearningCoefficientPto	TYPEDEF	 PROTO  :DWORD, :DWORD
NNet_backPropagatePto			TYPEDEF	 PROTO  :DWORD
NNet_runPto						TYPEDEF	 PROTO	:DWORD, :DWORD
NNet_getInfoPto					TYPEDEF  PROTO  :DWORD, :DWORD
NNet_savePto					TYPEDEF	 PROTO  :DWORD, :DWORD, :DWORD
NNet_loadPto					TYPEDEF  PROTO  :DWORD, :DWORD, :DWORD
NNet_getLastErrorPto			TYPEDEF  PROTO  :DWORD

; --=====================================================================================--
; CLASS STRUCTURE
; --=====================================================================================--
CLASS NNet
	; -------------------- <methods> ----------------------------
	CMETHOD destructor     	    	     ; MUST BE THE FIRST, OR OBJECTS.INC WILL FAIL
	CMETHOD train
	CMETHOD	initRandom
	CMETHOD calcOutputs
	CMETHOD calcOutputDeltas
	CMETHOD	setLearningCoefficient
	CMETHOD backPropagate
	CMETHOD run
	CMETHOD getInfo
	CMETHOD save
	CMETHOD load
	CMETHOD getLastError
	nInputs			dd		?
	nOutputs		dd		?
	nHiddenLayers	dd		?	
	nHiddenNeurons	dd		?	
	nLayerNeurons	dd		?	; max number of neurons on one layer (of all layers)
	
	lpOutputArr		dd		?	; Output array
	lpWeightArr		dd		?	; Weights array
	lpDeltaArr		dd		?	; Temporary delta array
	lnOutputArr		dd		?	; size of output array
	lnWeightArr		dd		?	; size of weight array
	lnDeltaArr		dd		?	; size of delta array
	
	lnLayerWeights	dd		?	; nr of weights for one layer transition
	
	r4learnCoeff	real4	?	; learning coefficient
	dwLastError		dd		?
NNet ENDS

include	NNet.inc
.data

BEGIN_INIT NNet
	dd offset NNet_destructor
	dd offset NNet_train
	dd offset NNet_initRandom
	dd offset NNet_calcOutputs
	dd offset NNet_calcOutputDeltas
	dd offset NNet_setLearningCoefficient
	dd offset NNet_backPropagate
	dd offset NNet_run
	dd offset NNet_getInfo
	dd offset NNet_save
	dd offset NNet_load
	dd offset NNet_getLastError
	dd 0,0,0,0,0	;n5x
	dd 0,0,0 		;lpOutputArr, lpWeightArr,lpDeltaArr
	dd 0,0,0		;lnOutputArr, lnWeightArr, lnDeltaArr
	dd 0			;lnLayerWeights
	real4 0.5		;r4learnCoeff
	dd 0			;
END_INIT NNet


flt32767	REAL4	32767.0


.code
;===============================================================================
; CServerThread::Init
;===============================================================================
NNet_Init proc  uses edi lpTHIS:DWORD, lpNNInfo:DWORD
   SET_CLASS NNet
  
  	SetObject edi, NNet
  	mov		ecx, lpNNInfo
  	assume	ecx:PTR NNINFO
  	
  	; --- Number of input neurons ---
  	mov		eax, [ecx].nInputs
  	mov		[edi].nInputs, eax
  	mov		edx, eax ;store #inputs in edx
  	
  	; --- Number of output neurons ---
  	mov		eax, [ecx].nOutputs
  	mov		[edi].nOutputs, eax
  	cmp		eax, edx
  	jbe		@F
  	mov		edx, eax ;store #outputs in edx if bigger than #inputs
  	@@:
  	
  	; --- Number of hidden neurons per layer ---
  	mov		eax, [ecx].nHiddenNeurons
  	mov		[edi].nHiddenNeurons, eax
  	cmp		eax, edx
  	jbe		@F
  	mov		edx, eax ;store #hidden in edx if bigger than #inputs & #outputs
  	@@:
  	
  	; --- Hidden layers ---
  	mov		eax, [ecx].nHiddenLayers
  	mov		[edi].nLayerNeurons, edx ;max neurons per layer (max of input,output and hidden)
  	mov		[edi].nHiddenLayers, eax
  	
  	assume	ecx:nothing
  	
  	; --- allocate memory for outputs ---
  	
  	; hiddenlayers + 2 = total nr of layers
  	add		eax, 2
  	inc		edx		;add one to max neurons per layer, for bias.
  	mul		edx		;total layers * (max neurons per layer + 1) = total output values
  	
  	shl		eax, 2 ; * 4 (dword entries)
  	mov		[edi].lnOutputArr, eax
  	invoke	GlobalAlloc, GMEM_FIXED OR GMEM_ZEROINIT, eax
  	mov		[edi].lpOutputArr, eax
  	
  	; --- allocate memory for weights ---
  	
  	mov		eax, [edi].nLayerNeurons
  	mov		ecx, eax
  	inc		ecx							;add one index for bias weight
  	mul		ecx		;(layerneurons+1) * (layerneurons)
  	mov		[edi].lnLayerWeights, eax
  	mov		ecx, [edi].nHiddenLayers
  	inc		ecx		;(hiddenlayers+1) = (totallayers-1)
  	mul		ecx  	;(totallayers-1)*(layerneurons+1)*(layerneurons)
  	shl		eax, 2 ; * 4 (dword entries)
  	mov		[edi].lnWeightArr, eax
  	invoke	GlobalAlloc, GMEM_FIXED, eax
  	mov		[edi].lpWeightArr, eax 
  	
  	; --- allocate memory for temp delta array ---
  	
  	mov		eax, [edi].nLayerNeurons
  	inc		eax ;add one for bias
  	shl		eax, 3 ; * 2 (current&next) * 4 (dword entries) = * 8
	invoke	GlobalAlloc, GMEM_FIXED OR GMEM_ZEROINIT, eax
	mov		[edi].lpDeltaArr, eax
  	
  	invoke	InitRand32
  	
	ReleaseObject edi
   ret
NNet_Init  endp

;===============================================================================
; NNet::destructor
;===============================================================================
NNet_destructor proc uses esi edi  lpTHIS:DWORD
   	SetObject edi, NNet
   	
   	invoke	GlobalFree, [edi].lpOutputArr
   	invoke	GlobalFree, [edi].lpWeightArr
   	invoke	GlobalFree, [edi].lpDeltaArr
	
	ReleaseObject edi
ret
NNet_destructor  endp
;.data
;tmp0302 REAL4 -1.0
;;tmp0303 REAL4 0.2
;;tmp0304	REAL4 0.341232, -0.115223, 0.129952, 0.570345, -0.923123,-0.328932
;;		REAL4 -0.993423,0.0,0.164732,0.0, 0.752621, 0.0,0.0,0.0
;
;
;
;
;.code

NNet_initRandom proc uses ebx esi lpTHIS:DWORD
LOCAL	 temp:DWORD
 
  	mov		eax, lpTHIS
  	mov		esi, (NNet ptr [eax]).lpWeightArr
  	mov		ebx, (NNet ptr [eax]).lnWeightArr
  	shr		ebx, 2
  	
  	.WHILE	ebx
  		invoke	RAND32, 32767*2
  		mov		temp, eax
		
		fild	temp
		fld		flt32767
		fsub	st(1),st(0)
		fdivp 	st(1),st(0)
		fstp	real4 ptr [esi]


  		dec		ebx
  		add		esi, 4
  	.ENDW
  	;temp:
;	mov		eax, lpTHIS
;	mov		eax, (NNet ptr [eax]).lpWeightArr
;	
;	invoke	MemCopy, addr tmp0304, eax, 12*4
	
ret
NNet_initRandom endp

;===============================================================================
; NNet::setLearningCoefficient
;===============================================================================
NNet_setLearningCoefficient proc lpTHIS:DWORD, learnCoeff:DWORD
 	mov		ecx, lpTHIS
 	mov		eax, learnCoeff
 	mov		(NNet ptr [ecx]).r4learnCoeff, eax
ret
NNet_setLearningCoefficient endp

;===============================================================================
; NNet::train
;===============================================================================
NNet_train proc uses edi lpTHIS:DWORD, lpInputs:DWORD, lpOutputs:DWORD
	SetObject edi, NNet

	; --- Copy inputs ---
	mov		eax, [edi].lpOutputArr
	add		eax, 4 ;skip bias value
	mov		ecx, [edi].nInputs
	shl		ecx, 2
	invoke	MemCopy, lpInputs, eax, ecx
	method	edi, NNet, calcOutputs
	method	edi, NNet, calcOutputDeltas, lpOutputs
	method	edi, NNet, backPropagate	
	
	ReleaseObject edi
ret
NNet_train endp

;===============================================================================
; NNet::calcOutputs
;===============================================================================
NNet_calcOutputs proc uses esi edi ebx lpTHIS:DWORD
LOCAL	lnLeft:DWORD	
LOCAL	lnRight:DWORD
LOCAL	lpCurLayerWeights:DWORD
LOCAL	dwCurLayer:DWORD
LOCAL	dwWeightSkipVal:DWORD
LOCAL 	dwOutputSkipVal:DWORD

	finit
	SetObject edi, NNet
	
	
	mov		eax, [edi].nLayerNeurons
	shl		eax, 2
	mov		dwWeightSkipVal, eax  ; bytes to skip for next source neuron
	
	add		eax, 4
	mov		dwOutputSkipVal, eax  ; bytes to skip for next layer output
	
	mov		eax, [edi].lpWeightArr
	mov		lpCurLayerWeights, eax
	
	mov		esi, [edi].lpOutputArr	
	; dwCurLayer: current transition calculated: dwCurLayer=0 means transition from layer1 to 2.
	and		dwCurLayer, 0
	jmp		@layerloop_end
	@layerloop:
		; ecx = dwCurLayer here
		
		; get number of neurons in current left layer
		.IF		ecx==0
			mov		edx, [edi].nInputs
		.ELSE
			mov		edx, [edi].nHiddenNeurons
		.ENDIF
		inc		edx ;bias
		mov	lnLeft, edx
		
		; get number of neurons in current right layer
		.IF		ecx==[edi].nHiddenLayers
			mov		edx, [edi].nOutputs
		.ELSE
			mov		edx, [edi].nHiddenNeurons
		.ENDIF
		inc		edx ;bias
		mov	lnRight, edx
		

		mov		edx, [edi].nLayerNeurons
		inc		edx ;add bias
		shl		edx, 2
		add		edx, esi
		
		; esi: current layer output data
		; edx: next layer output data
		
		; --- loop through all neurons in next (right) layer ---
		xor		ebx, ebx
		inc		ebx			;bias is not outputted
		.WHILE	ebx<lnRight
			mov		eax, ebx	; (i - 1)*4 
			dec		eax			
			shl		eax, 2
			add		eax, lpCurLayerWeights
			
			mov		dword ptr [esi], 3F800000h ; 1.0 as dword, bias value
	
			fld1
			fldz
			xor		ecx, ecx
			.WHILE	ecx<lnLeft
				fld		dword ptr [eax]		  ;load weight value
				fmul 	dword ptr [esi+ecx*4] ;load input value
				add		eax, dwWeightSkipVal
				faddp	st(1), st(0)
			inc ecx
			.ENDW
			fchs
			EXPVAL
			fadd	st(0),st(1)
			fdivp	st(1),st(0)
			
			fstp	real4 ptr [edx+ebx*4]
			
		inc		ebx
		.ENDW
	
		mov		eax, [edi].lnLayerWeights
		shl		eax, 2
		add		lpCurLayerWeights, eax
		add		esi, dwOutputSkipVal
		inc		dwCurLayer
	@layerloop_end:
	mov		ecx, dwCurLayer
	cmp		ecx, [edi].nHiddenLayers
	jle		@layerloop
		
	
	ReleaseObject edi
ret
NNet_calcOutputs endp


;===============================================================================
; NNet::calcOutputDeltas
;===============================================================================
NNet_calcOutputDeltas proc uses esi edi ebx lpTHIS:DWORD, lpOutputs:DWORD
	SetObject edi, NNet
	; delta = y . [1 - y] . [d - y]
	mov		esi, [edi].nLayerNeurons
	inc		esi 	;account for bias value
	shl		esi, 2 
	neg		esi
	add		esi, [edi].lpOutputArr
	add		esi, [edi].lnOutputArr
	add		esi, 4 ; skip bias
	
	mov		ebx, lpOutputs
	mov		edx, [edi].lpDeltaArr
	
	; esi: points to produced output array (index 1, bias value skipped)
	; ebx: points to desired output array
	; edx: points to delta array.
	finit
	xor		eax, eax
	mov		ecx, [edi].nOutputs
	.WHILE	eax<ecx
		fld		real4 ptr [esi+eax*4]		;load y
		fld		real4 ptr [ebx+eax*4]		;load d
		fsub	st(0), st(1)				;st(0)=d-y, st(1)=y
		fld1						
		fsub	st(0), st(2)				;st(0)=1-y, st(1)=d-y, st(2)=y
		fmulp	st(1), st(0)				;st(0)=[d-y]*[1-y], st(1)=y
		fmulp	st(1), st(0)				;st(0)=y*[1-y]*[d-y]=delta
		fstp	real4 ptr [edx+eax*4+4]		;store delta (+4 to skip delta)
		inc		eax
	.ENDW
	
	ReleaseObject edi
ret	
NNet_calcOutputDeltas endp

;===============================================================================
; NNet::backPropagate
;===============================================================================
NNet_backPropagate proc uses esi edi ebx lpTHIS:DWORD
LOCAL	lnLeft:DWORD	
LOCAL	lnRight:DWORD
LOCAL	lpCurLayerWeights:DWORD
LOCAL	dwCurLayer:DWORD
LOCAL	dwWeightSkipVal:DWORD
LOCAL 	dwOutputSkipVal:DWORD
LOCAL	lpCurOutputs:DWORD
LOCAL	bSwapFlag:DWORD
	finit
	SetObject edi, NNet

	and		bSwapFlag, 0
	
	mov		eax, [edi].nLayerNeurons
	shl		eax, 2
	mov		dwWeightSkipVal, eax  ; bytes to skip for prev source neuron
		
	add		eax, 4
	mov		dwOutputSkipVal, eax  ; bytes to skip for prev layer output
	
	mov		eax, [edi].lpWeightArr
	add		eax, [edi].lnWeightArr
	mov		ecx, [edi].lnLayerWeights
	shl		ecx, 2
	sub		eax, ecx	;move to last layer transition
	mov		lpCurLayerWeights, eax
	
	
	mov		eax, [edi].nLayerNeurons
	inc		eax 	;account for bias value
	shl		eax, 3  ; * 4 for dword indexes, * 2 for two layers == * 8 
	neg		eax
	add		eax, [edi].lpOutputArr
	add		eax, [edi].lnOutputArr
	mov		lpCurOutputs, eax
	
	
	; dwCurLayer: current transition calculated: dwCurLayer=0 means transition from layer1 to 2.
	mov		eax, [edi].nHiddenLayers
	inc		eax
	mov		dwCurLayer, eax
	jmp		@layerloop_end
	@layerloop:
		dec		ecx
		; ecx = dwCurLayer here

		; get number of neurons in current left layer
		.IF		ecx==0
			mov		edx, [edi].nInputs
		.ELSE
			mov		edx, [edi].nHiddenNeurons
		.ENDIF
		inc		edx ;bias
		mov	lnLeft, edx
		
		; get number of neurons in current right layer
		.IF		ecx==[edi].nHiddenLayers
			mov		edx, [edi].nOutputs
		.ELSE
			mov		edx, [edi].nHiddenNeurons
		.ENDIF
		inc		edx ;bias
		mov	lnRight, edx
		

		mov		esi, [edi].lpDeltaArr
		mov		edx, esi
		.IF		bSwapFlag
			add		esi, dwOutputSkipVal
		.ELSE
			add		edx, dwOutputSkipVal
		.ENDIF

		mov		eax, lpCurLayerWeights
		
		; esi: points to delta array of right layer [in]
		; edx: points to delta array of left layer [out]
		; eax: points to current layer weights array [in]
		
		; --- loop through all neurons in current (left) layer ---
		push	edi

		mov		edi, lpCurOutputs
		; edi: points to output array of left layer [in]

		; --- calculate left layer deltas loop ---
		xor		ebx, ebx
		.WHILE	ebx<lnLeft
			xor		ecx, ecx
			fldz			;start with zero for SUM
			inc		ecx		;skip bias
			.WHILE	ecx<lnRight
				fld	  real4 ptr [eax+ecx*4-4] 	; load weight value (-4 is for the 1 in (i-1))
				fmul  real4	ptr	[esi+ecx*4]		; multiply by delta value of right neuron
				faddp st(1),st(0)				; add to current sum
			inc	ecx
			.ENDW
			; out . [1 - out]
			;!!!!!!!!!!!!!!!!!! bugfix in next line, ebx was ecx:
			fld		real4 ptr [edi+ebx*4]		; load output value of left neuron
			fld1	
			fsub	st(0), st(1)
			fmulp	st(1), st(0)
			fmulp	st(1), st(0)				; out . [1-out]
			fstp	real4 ptr [edx+ebx*4]		; store delta value for neuron in left layer
			add		eax, dwWeightSkipVal
		inc		ebx
		.ENDW

		mov		edx, edi		;edx takes over edi (ptr output array of left layer)
		pop		edi
		
		mov		eax, lpCurLayerWeights
		
		; esi: points to delta array of right layer [in]
		; edx: points to output array of left layer [in]
		; eax: points to current layer weights array [out]
	
		fld		[edi].r4learnCoeff			   ; load learning coefficient
		; --- adjust weights loop ---
		xor		ebx, ebx
		.WHILE	ebx<lnLeft
			xor		ecx, ecx
			inc		ecx		;skip bias
			.WHILE	ecx<lnRight
				fld		real4 ptr [edx+ebx*4]	; load output of left neuron
				fmul  	real4 ptr [esi+ecx*4]	; multiply by delta value of right neuron 
				fmul	st(0), st(1)			; multiply by learning coefficient
				fadd	real4 ptr [eax+ecx*4-4] ; add delta to old weight value (-4 is for the 1 in (i-1))
				fstp	real4 ptr [eax+ecx*4-4] ; store new delta value				
			inc	ecx
			.ENDW
			add		eax, dwWeightSkipVal
		inc		ebx
		.ENDW
		
		fstp	st(0)							; remove learning coefficient
		
		mov		eax, [edi].lnLayerWeights
		shl		eax, 2
		sub		lpCurLayerWeights, eax
		mov		eax, dwOutputSkipVal
		xor		bSwapFlag, 1

		sub		lpCurOutputs, eax
		dec		dwCurLayer
	@layerloop_end:
	mov		ecx, dwCurLayer
	or		ecx, ecx
	jnz		@layerloop
	ReleaseObject edi
ret	
NNet_backPropagate endp


;===============================================================================
; NNet::run
;===============================================================================
NNet_run proc uses edi lpTHIS:DWORD, lpInputs:DWORD
	SetObject edi, NNet

	; --- Copy inputs ---
	mov		eax, [edi].lpOutputArr
	add		eax, 4 ;skip bias value
	mov		ecx, [edi].nInputs
	shl		ecx, 2
	invoke	MemCopy, lpInputs, eax, ecx
	method	edi, NNet, calcOutputs

	mov		eax, [edi].nLayerNeurons
	inc		eax
	shl		eax, 2
	neg		eax
	add		eax, [edi].lpOutputArr
	add		eax, [edi].lnOutputArr
	add		eax, 4
	ReleaseObject edi
ret
NNet_run endp

;===============================================================================
; NNet::getInfo
;===============================================================================
NNet_getInfo proc uses edi lpTHIS:DWORD, dwType:DWORD
	SetObject edi, NNet
	and		[edi].dwLastError, 0
	
	mov		eax, dwType
	.IF		eax==NNETINFO_MEMUSE
		mov		eax, [edi].lnWeightArr
		add		eax, [edi].lnDeltaArr
		add		eax, [edi].lnOutputArr
	.ELSEIF eax==NNETINFO_TOTALNEURONS
		mov		eax, [edi].nHiddenLayers
		mul		[edi].nHiddenNeurons
		add		eax, [edi].nInputs
		add		eax, [edi].nOutputs
	.ELSEIF	eax==NNETINFO_SERDATASIZE
		mov		eax, [edi].lnWeightArr
		add		eax, SIZEOF NNETSERIAL_HEADER
	.ELSE
		mov		[edi].dwLastError, NNETERR_INVALIDPARAM
		or	eax, -1
	.ENDIF
	ReleaseObject edi
ret
NNet_getInfo endp

;===============================================================================
; NNet::save
;===============================================================================
NNet_save proc uses edi lpTHIS:DWORD, dwFlags:DWORD, outData:DWORD
LOCAL netHeader:NNETSERIAL_HEADER
LOCAL bytesWritten:DWORD
	SetObject edi, NNet
	and		[edi].dwLastError, 0
	
	mov		netHeader.dwMagic, "NNET"
	mov 	eax, [edi].lnWeightArr
	mov	 	netHeader.lnWeights, eax
	
	mov		ecx, dwFlags
	.IF		(ecx & NNETFLAG_TOFILE)
		invoke	WriteFile, outData, addr netHeader, SIZEOF netHeader, addr bytesWritten, 0
		.IF		!eax
			mov	[edi].dwLastError, NNETERR_FILEACCESS
			xor	eax, eax
			ret
		.ENDIF
		invoke	WriteFile, outData, [edi].lpWeightArr, [edi].lnWeightArr, addr bytesWritten, 0
		.IF		!eax
			mov	[edi].dwLastError, NNETERR_FILEACCESS
			xor	eax, eax
			ret
		.ENDIF
		xor		eax, eax
		inc	 	eax
	.ELSEIF (ecx & NNETFLAG_TOMEM)
		invoke	MemCopy, addr netHeader, outData, SIZEOF netHeader 
		mov		edx, outData
		add		edx, SIZEOF netHeader
		invoke	MemCopy, [edi].lpWeightArr, outData, [edi].lnWeightArr
		xor		eax, eax
		inc	 	eax
	.ELSE
		mov		[edi].dwLastError, NNETERR_INVALIDPARAM
		xor		eax, eax
	.ENDIF
	ReleaseObject edi
ret
NNet_save endp

;===============================================================================
; NNet::load
;===============================================================================
NNet_load proc uses edi lpTHIS:DWORD, dwFlags:DWORD, inData:DWORD
LOCAL netHeader:NNETSERIAL_HEADER
LOCAL bytesRead:DWORD
	SetObject edi, NNet
	and		[edi].dwLastError, 0

	mov		ecx, dwFlags
	.IF		(ecx & NNETFLAG_FROMFILE)
		invoke	ReadFile, inData, addr netHeader, SIZEOF netHeader, addr bytesRead, 0
		.IF		(!eax) || (bytesRead!=SIZEOF netHeader)
			mov	[edi].dwLastError, NNETERR_FILEACCESS
			xor	eax, eax
			jmp	@return
		.ENDIF
		mov		eax, netHeader.lnWeights
		.IF		netHeader.dwMagic!="NNET" || eax!=[edi].lnWeightArr
			mov	[edi].dwLastError, NNETERR_INVALIDDATA
			xor	eax, eax
			jmp	@return
		.ENDIF
		invoke	ReadFile, inData, [edi].lpWeightArr, [edi].lnWeightArr, addr bytesRead, 0
		mov		ecx, netHeader.lnWeights
		.IF		(!eax) || (bytesRead!=ecx)
			mov	[edi].dwLastError, NNETERR_FILEACCESS
			xor	eax, eax
			jmp	@return
		.ENDIF
		xor		eax, eax
		inc	 	eax
	.ELSEIF (ecx & NNETFLAG_FROMMEM)
		mov		edx, inData
		assume	edx:PTR NNETSERIAL_HEADER
		mov		eax, [edx].lnWeights
		.IF		[edx].dwMagic!="NNET" || eax!=[edi].lnWeightArr
			mov	[edi].dwLastError, NNETERR_INVALIDDATA
			xor	eax, eax
			jmp	@return
		.ENDIF
		assume	edx:nothing
		add		edx, 8
		invoke	MemCopy, edx, [edi].lpWeightArr, [edi].lnWeightArr
		xor		eax, eax
		inc		eax
	.ELSE
		mov		[edi].dwLastError, NNETERR_INVALIDPARAM
		xor		eax, eax
	.ENDIF
@return:
	ReleaseObject edi
ret
NNet_load endp

;===============================================================================
; NNet::getLastError
;===============================================================================
NNet_getLastError proc lpTHIS:DWORD
	mov		eax, lpTHIS
	mov		eax, (NNet ptr [eax]).dwLastError
ret
NNet_getLastError endp